/***************************************************************************
 *                          Zanoza Modeler v2.x                            *
 *      This is unpublished propietary source code of Zanoza Software.     *
 *      The copyright notice does not evidence any actual or intended      *
 *                    publication of such source code.                     *
 *                                                                         *
 *                   Copyright (c) 2002 Zanoza Software.                   *
 *                                                                         *
 *                        All rights reserved.                             *
 ***************************************************************************
 *--------------------------------------------------------------------------
 * @author:   Oleg M.
 *--------------------------------------------------------------------------
 * @purpose:  ZModeler 2.x CSplineNode shared class implementation
 *--------------------------------------------------------------------------
 * @histroy: 17.09.2003. Created
 */
#include "Spline.h"
#include "SplineNode.h"
#include "SplineTechnique.h"
#include <helpers/events.h>
#include <helpers/sceneNavigation.h>
#include <core/ui/ISceneViewport.h>

namespace scene
{

INTERFACE_BEGINMAP(CSplineNode)
  INTERFACE_ENTRY(scene::INode)
  INTERFACE_ENTRY(scene::ISplineNode)
  INTERFACE_ENTRY(scene::ITransformable)
  INTERFACE_ENTRY(scene::IVisualizeable)
  INTERFACE_ENTRY(scene::IBoundable)
  INTERFACE_ENTRY(scene::IManipulateable)
  INTERFACE_ENTRY(core::io::ISerializeable)
  INTERFACE_ENTRY(core::INamed)
  INTERFACE_EMBED(core::ui::controls::IUserDefinedOptions, m_pUserOptions.m_p)
  INTERFACE_EMBEDDED_CHAIN(scene::CMiniNode, m_pNode.m_p)
INTERFACE_ENDMAP()



CSplineNode::CSplineNode() :  CVisualizeable(), 
                              CBoundable(),
                              core::CNamed(),
                              scene::CTransformable()
{
  m_dwStatus = STATUS_ENABLED;
  CManipulatorCol::CreateInstance(&m_pControllers);
  m_strName = m_strTitle = _T("Spline");
}

CSplineNode::~CSplineNode()
{
}

void CSplineNode::FinalConstruct()
{
  CMiniNode::CreateOwnedInstance(&m_pNode, GetControllingUnknown());
  core::ui::controls::CUserDefinedOptions::CreateOwnedInstance(&m_pUserOptions, GetControllingUnknown());
}


//-------------------------------------------------------
// @name : CSplineNode::getSpline
// Provides Spline associated with the node
// @param scene::ISpline** ppSpline          : ret-val arg
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSplineNode::getSpline(scene::ISpline** ppSpline)
{
  return m_pSpline.CopyTo(ppSpline);
}


//-------------------------------------------------------
// @name : CSplineNode::setSpline
// Sets mesh for the node
// @param scene::ISpline* pSpline            : new Spline for the node
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSplineNode::setSpline(scene::ISpline* pSpline)
{
  m_pSpline = pSpline;
  if (!m_pSpline)
    return ZRESULT_FALSE;
  return changed(NC_VISUAL | NC_GEOMETRY);
}


//-------------------------------------------------------
// @name : CSplineNode::queryNodeOperation
// Child operations on Spline node
// @param eNodeOperation co             : 
// @param scene::INode* pNode            : 
// @param DWORD dwRes                    : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSplineNode::queryNodeOperation(eNodeOperation co, scene::INode* pNode, DWORD dwRes)
{
  return ZRESULT_OK;//allow everything
}

//-------------------------------------------------------
// @name : CSplineNode::changed
// Mesh node's changed.
// @param bool bGeometry                 : 
// @param bool bVisual                   : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSplineNode::changed(DWORD dwBits)
{
  if (!m_pSpline)
    return ZRESULT_OK;
  m_pSpline->pushState();
  //
  // alter boundary and manipulators
  //
  if (dwBits & NC_GEOMETRY)
  {
    updateBoundary();
    if (m_pControllers && m_pSpline)
    {
      long numManips = 0;
      scene::ISpline::lpNode ptr = m_pSpline->head();
      while (ptr)
      {
        numManips++;//position
        scene::eControlPoint type;
        ptr->m_pIntr->getType(type);
        if ((ZRESULT_OK == m_pSpline->isClosed() || ptr != m_pSpline->head())
            && scene::SCP_CORNER != type)
          numManips++;//incoming
        if ((ZRESULT_OK == m_pSpline->isClosed() || NULL != ptr->m_pNext)
            && scene::SCP_CORNER != type)
          numManips++;//outgoing
        ptr = ptr->m_pNext;
      }
      if (m_pControllers->length() != numManips)
        updateManipulators();
    }
    //ShowMessageDirect(core::MT_DEBUG, _T("spline changed"));
  }
  if (m_pRenderTechnique)
    m_pRenderTechnique->nodeChange(this, dwBits);
  m_pSpline->popState();
  return ZRESULT_OK;
}


//-------------------------------------------------------
// @name : CSplineNode::setStatus
// Status-specific stuff
// @param DWORD nBits                    : 
//
// @return ZRESULT : 
//-------------------------------------------------------
DWORD CSplineNode::setStatus(const DWORD nBits)
{
  DWORD dwAdd     = nBits & ~ZEntityBase::getStatus();
  DWORD dwRemove  = ZEntityBase::getStatus() & ~nBits;
  DWORD dwChange = (dwAdd | dwRemove);
  DWORD dw = ZEntityBase::setStatus(nBits);
  if (dwChange & (STATUS_ACTIVE | STATUS_SELECTED | STATUS_HIDDEN))
    if (m_pRenderTechnique)
      m_pRenderTechnique->nodeChange(this, NC_VISUAL);
  return dw;
}

DWORD CSplineNode::changeStatus(const DWORD nAdd, const DWORD nRemove)
{
  DWORD dwChange = (nAdd | nRemove);
  DWORD dw = ZEntityBase::changeStatus(nAdd, nRemove);
  if (dwChange & (STATUS_ACTIVE | STATUS_SELECTED | STATUS_HIDDEN))
    if (m_pRenderTechnique)
      m_pRenderTechnique->nodeChange(this, NC_VISUAL);
  return dw;
}



//-------------------------------------------------------
// @name : CSplineNode::clone
// Creates a copy of spline
// @param scene::INode** ppNode          : 
// @param bool bShareData                : 
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSplineNode::clone(scene::INode** ppNode, bool bShareData)
{
  if (NULL == ppNode)
    return ZRESULT_FALSE;
  ZPtr<scene::ISplineNode>      ptr;
  ZPtr<scene::IScene>           pScene;
  ZPtr<services::INodeService>  pService;
  g_pZModeler->getScene(&pScene);
  if (pScene && ZRESULT_OK == pScene->getNodeService(&pService))
    pService->createNode(IID_(scene::ISplineNode), (void**)&ptr);
  else
    scene::CSplineNode::CreateInstance((void**)&ptr);
  if (!ptr)
    return ZRESULT_FALSE;

  if (m_pNode)
  {
    ZString     str;
    ZStringSeq  seq;
    ZBoundary   bound;
    ZMatrix     mat;
    ZPtr<core::INamed>          pNamed;
    ZPtr<scene::ITransformable> pTrans;
    getTitle(str);
    ptr->QueryInterface(IID_(core::INamed), (void**)&pNamed);
    if (pNamed)
      pNamed->setTitle(str);
    getName(str);
    if (pNamed)
      pNamed->setName(str);
    getDescription(seq);
    if (pNamed)
      pNamed->setDescription(seq);
    getTransform(mat);
    ptr->QueryInterface(IID_(scene::ITransformable), (void**)&pTrans);
    if (pTrans)
      pTrans->setTransform(mat);
  }
  if (bShareData)
    ptr->setSpline(m_pSpline);
  else
  {
    //
    // create a copy of Spline...
    //
    ZPtr<scene::CSpline>          pSpline;
    CSpline::CreateInstance(&pSpline);
    ZPtr<scene::IControlPoint>    pt;
    pSpline->setClosed(ZRESULT_OK == m_pSpline->isClosed());
    m_pSpline->reset();
    ZPoint3D pos;
    ZVector  v1,v2;
    scene::eControlPoint type;
    while (ZRESULT_OK == m_pSpline->next(&pt))
    {
      pt->getControlPosition(pos);
      pt->getType(type);
      pt->getControlVector(SCV_U0, v1);
      pt->getControlVector(SCV_U1, v2);
      ZPtr<scene::CControlPoint>    ptNew;
      scene::CControlPoint::CreateInstance(&ptNew);
      ptNew->setControlPosition(pos);
      ptNew->setType(type);
      ptNew->setControlVector(SCV_U0, v1);
      ptNew->setControlVector(SCV_U1, v2);
      pSpline->addLast(ptNew);
      pt = NULL;
    }
    //create manipulators
    updateManipulators();

    //done. set the spline:
    ptr->setSpline(pSpline);
  }
  ptr->setParent(NULL); // unattached
  //
  // copy user-defined options, if available
  ZPtr<core::ui::controls::IUserDefinedOptions> pUserOpt;
  if (m_pUserOptions && 
      ZRESULT_OK == ptr->QueryInterface(IID_(core::ui::controls::IUserDefinedOptions), (void**)&pUserOpt))
      core::ui::controls::copyUserDefinedOptions(m_pUserOptions, pUserOpt);

  ptr->changed(NC_VISUAL | NC_GEOMETRY);
  (*ppNode) = ptr;
  (*ppNode)->AddRef();
  return ZRESULT_OK;
}



//-----------------------
// IManipulateable
//-----------------------
ZRESULT CSplineNode::getManipulators(CManipulatorCol** ppPoints)
{
  return m_pControllers.CopyTo(ppPoints);
}

ZRESULT CSplineNode::getManipulatorsByName(ZString strName, CManipulatorCol** ppPoints)
{
  return ZRESULT_NOT_IMPLEMENTED;
}


//-------------------------------------------------------
// @name : clampMinMax
// Helper to clamp min and max
// @param ZPoint3D& pt                   : 
// @param ZPoint3D& ptMin                : 
// @param ZPoint3D& ptMax                : 
//
// @return void : 
//-------------------------------------------------------
void clampMinMax(ZPoint3D& pt, ZPoint3D& ptMin, ZPoint3D& ptMax)
{
  if (ptMin.x > pt.x)
    ptMin.x = pt.x;
  if (ptMin.y > pt.y)
    ptMin.y = pt.y;
  if (ptMin.z > pt.z)
    ptMin.z = pt.z;
  if (ptMax.x < pt.x)
    ptMax.x = pt.x;
  if (ptMax.y < pt.y)
    ptMax.y = pt.y;
  if (ptMax.z < pt.z)
    ptMax.z = pt.z;
}

//-----------------------
// IBoundable
//-----------------------
ZRESULT CSplineNode::updateBoundary()
{
  scene::ISpline::lpNode ptr = m_pSpline ? m_pSpline->head() : NULL;
  scene::ISpline::lpNode ptr2;
  if (NULL == ptr)
    return ZRESULT_FALSE;
  float fMAX = 1.0f/MINFLOAT;
  ZPoint3D pt, ptMin, ptMax;
  ptMin.x = ptMin.y = ptMin.z = fMAX;
  ptMax.x = ptMax.y = ptMax.z =-fMAX;
  while (ptr)
  {
    if (ptr)
      ptr2 = ptr->m_pNext;
    if (NULL == ptr2 && ZRESULT_OK == m_pSpline->isClosed())
      ptr2 = m_pSpline->head();

    ZPoint3D pt;
    long nDetail;
    ptr->m_pIntr->getDetalization(nDetail);
    ptr->m_pIntr->getControlPosition(pt);
    clampMinMax(pt, ptMin, ptMax);
    if (ptr2)
      for (int i = 0; i < nDetail+1; i++)
      {
        //get point
        float u = (float)(i)/(float)(nDetail+1);
        m_pSpline->getSectionPoint(ptr->m_pIntr, ptr2->m_pIntr, u, pt);
        clampMinMax(pt, ptMin, ptMax);
      }
    ptr = ptr->m_pNext;
    ptr2 = NULL;
  };

  m_Bound.setBoxBound(ptMin, ptMax);
  //m_Bound.setSphereBound((ptMin + ptMax)/2, 0.5f*(ptMax-ptMin).length());
  return ZRESULT_OK;
}



//-------------------------------------------------------
// @name : CSplineNode::updateManipulators
// Updates manipulators
//
// @return ZRESULT : 
//-------------------------------------------------------
ZRESULT CSplineNode::updateManipulators()
{
  m_pControllers->clear();
  scene::ISpline::lpNode ptr = m_pSpline ? m_pSpline->head() : NULL;
  if (NULL == ptr)
    return ZRESULT_FALSE;
  ZPoint3D pt;
  ZVector  v1,v2;
  scene::eControlPoint type;
  ZPtr<CPositionManipulator>    pPos;
  ZPtr<CDirectionManipulator>   pV1, pV2;
  while (ptr)
  {
    ptr->m_pIntr->getControlPosition(pt);
    ptr->m_pIntr->getType(type);
    ptr->m_pIntr->getControlVector(SCV_U0, v1);
    ptr->m_pIntr->getControlVector(SCV_U1, v2);
    //create manipulators
    CPositionManipulator::CreateInstance(&pPos);
    pPos->init(this, ptr->m_pIntr);

    if ((ZRESULT_OK == m_pSpline->isClosed() || ptr != m_pSpline->head())
        && scene::SCP_CORNER != type)
    {
      CDirectionManipulator::CreateInstance(&pV2);
      pV2->init(this, ptr->m_pIntr, scene::SCV_U1);
      m_pControllers->addLast(pV2);
    }
    if ((ZRESULT_OK == m_pSpline->isClosed() || NULL != ptr->m_pNext)
        && scene::SCP_CORNER != type)
    {
      CDirectionManipulator::CreateInstance(&pV1);
      pV1->init(this, ptr->m_pIntr, scene::SCV_U0);
      m_pControllers->addLast(pV1);
    }
    m_pControllers->addLast(pPos);//pos is the last (for easier picking of vectors)
    pPos = NULL;
    pV1 = NULL;
    pV2 = NULL;
    ptr = ptr->m_pNext;
  }
  //::ShowMessageDirect(core::MT_DEBUG, _T("Spline::UpdateManipulators: %li manips"), m_pControllers->length());
  return ZRESULT_OK;
}

//-------------------------------------------------------
// @name : getSpline
// Provides a Spline from supplied node (if available)
// @param INode* pNode                   : 
// @param ISpline** ppSpline               : 
//
// @return bool : 
//-------------------------------------------------------
bool getSpline(INode* pNode, ISpline** ppSpline)
{
  if (NULL == pNode || NULL == ppSpline)
    return false;
  ZPtr<ISplineNode> pSplineNode;
  if (ZRESULT_OK != pNode->QueryInterface(IID_(scene::ISplineNode), (void**)&pSplineNode))
    return false;
  return ZRESULT_OK == pSplineNode->getSpline(ppSpline);
}


//-------------------------------------------------------
//
// manipulators
//
//-------------------------------------------------------
CPositionManipulator::CPositionManipulator()
{
}

CPositionManipulator::~CPositionManipulator()
{
}

//-------------------------------------------------------
// @name : CPositionManipulator::init
// Initialization
// @param scene::INode* pNode            : 
// @param IControlPoint* pControl        : 
//
// @return void : 
//-------------------------------------------------------
void CPositionManipulator::init(scene::INode* pNode, IControlPoint* pControl)
{
  m_pParent = pNode;
  m_pControl = pControl;
}

void CPositionManipulator::getPosition(ZPoint3D& pt)
{
  if (m_pControl)
    m_pControl->getControlPosition(pt);
}

void CPositionManipulator::setPosition(ZPoint3D& pt)
{
  if (m_pControl)
    m_pControl->setControlPosition(pt);
  m_pParent->changed(NC_CHILD | NC_VISUAL);
}

CDirectionManipulator::CDirectionManipulator()
{
}

CDirectionManipulator::~CDirectionManipulator()
{
}

void CDirectionManipulator::init(scene::INode* pNode, IControlPoint* pControl, eControlVector type)
{
  m_pParent = pNode;
  m_pControl = pControl;
  m_type = type;
}

void CDirectionManipulator::getPosition(ZPoint3D& pt)
{
  if (m_pControl)
  {
    m_pControl->getControlPosition(pt);
    ZVector v;
    m_pControl->getControlVector(m_type, v);
    pt = pt + v;
  }
}

void CDirectionManipulator::setPosition(ZPoint3D& pt)
{
  if (m_pControl)
  {
    ZPoint3D p;
    m_pControl->getControlPosition(p);
    ZVector v = pt-p;
    m_pControl->setControlVector(m_type, v);
    m_pParent->changed(NC_CHILD | NC_VISUAL);
  }
}


};//scene

